
/* amosdio.  
	Handle plain bit-wise IO on the DIO connector on the AMOS-3000
	There are 2 open-drain lines that may be used as inputs or outputs,
	and 2 input-only lines. One of these is set to hit the system with
	an SCI, this linkage can be optionally turned off here. 

	Read and write is one byte each:
Bit #: 7 6 5 4  3     2    1    0      -
Read:  0 1 0 0 GPI5 GPI4 GPIO3 GPIO2   - 
Write: X X X X  X     X  GPIO3 GPIO2   - 

ioctls: 

Ideas further on: 
	i2c protocol on the two bidirectional lines, linking into
	the i2c stack using i2c-dev and so on.
	Support some kind of external shift-register action	
*/ 

#include <linux/init.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/major.h>
#include <linux/module.h>
#include <linux/pci.h>

#include <linux/kernel.h>
#include <linux/ioctl.h>

#include <linux/ioport.h>
#include <linux/errno.h>
#include <asm/uaccess.h>
#include <asm/io.h> 

#include "amosdio.h" 

#define AMOSDIONAME "amosdio"
#define AMOSDIOMAJOR 61

int amosdio_open(struct inode *inode, struct file *filp);
int amosdio_release(struct inode *inode, struct file *filp);
ssize_t amosdio_read(struct file *filp, char *buff, size_t count, loff_t *offp);
ssize_t amosdio_write(struct file *filp, const char *buff, size_t count, loff_t *offp);
int amosdio_ioctl(struct inode *inode, struct file *filp, unsigned int code, unsigned long param);



/* Set these to make the system shut up about "tainting" of the kernel. 
	Since I made this thing, the license itself is rather irrelevant.  */
MODULE_DESCRIPTION("Simple Driver for the DIO connection on the AMOS-3000"); 
MODULE_AUTHOR("Knut Reidar Leer");
MODULE_LICENSE("GPL");

struct file_operations amosdio_fops = 
{
	.read = amosdio_read,
	.write = amosdio_write,
	.ioctl = amosdio_ioctl, 
	.open = amosdio_open,
	.release = amosdio_release,
	.owner = THIS_MODULE,
};

/* 
We want to look at PCI bus 0 device 17 function 0
	register 0x81, that bit 7 there is set. If it isn't we
	don't have access...

Then we want to read the io port base address from 
	PCI bus 0 device 17 function 0 register 0x88 and 0x89
	as a 16-bit quantity.

The bits have positive logic (1=H 0=L) and the chenge-flags
are reset by writing 1 to the same bit. Outputs are open drain, 
so setting them at H (1) makes them useable as inputs. 

The GPIO2-GPIO3 are then found at these ports:
	Input state: +0x49 bit 6-bit 7
    Change flag: +0x50 bit 6-bit 7 set on changes either direction
	Output bit: +0x4d  bit 6-bit 7

For GPI4-GPI5:
	Input state: +0x48 bit 4-bit 5
    Change flag: +0x50 bit 4-bit 5 set on change H-L 
	+0x22 bit 4 is set making this initiate a SMI and suspend action.
	Have this controllable in the ioctl function.
*/ 
	
static int amosdio_lo_ucount;
static unsigned short amosdio_PMIObase; 

/* As it turns out, the ports 
	0x420-0x423 are used by ACPI GPE0_BLK
	0x450-0x453 are used by ACPI GPE1_BLK
But ports 0x448, 0x449, 0x44d are not taken, so 
I call dibs on these! */ 

int amosdio_init(void)
{
	int erx;
	struct resource *AR1; 
	struct resource *AR2; 
	struct pci_dev *pdev; 
	unsigned char value8; 

	pdev = pci_get_bus_and_slot(0, PCI_DEVFN(17, 0)); 
	// printk("<1>%s(%d): bus 0 device 17 function 0 device %p\n", "amosdio.c", __LINE__, pdev); 
	if(!pdev)
	{
		printk("<1>%s(%d): Finding anything at bus 0 device 17 function 0 failed.\n", "amosdio.c", __LINE__); 
		return(-1);
	}

	pci_bus_read_config_byte(pdev->bus, PCI_DEVFN(17,0), 0x81, &value8);
	if( (value8 & 0x80) == 0)
	{
		pci_dev_put(pdev); 
		printk("<1>%s(%d): B0D17F0 Rx81=0x%x Bit 7 clear; No access to PMIO\n", 
			"amosdio.c", __LINE__, value8); 
		return(-1);
	}
	printk("<1>%s(%d): B0D17F0 Rx81=0x%x Bit 7 set; Access to PMIO is oke\n", 
			"amosdio.c", __LINE__, value8); 

	/* now get the base-address */ 	
	pci_bus_read_config_word(pdev->bus, PCI_DEVFN(17,0), 0x88, &amosdio_PMIObase);
	amosdio_PMIObase &= 0xFF80; 
	// printk("<1>%s(%d): PMIO base is at 0%x\n", "amosdio.c", __LINE__, amosdio_PMIObase); 

	/* We now release the PCI device after we're done with it. */ 
	/* Further accesses are in the IO-space at PMIObase + offsets */ 
	pci_dev_put(pdev); 

	/* Reserve or hit the AMOSDIO IO ports of interest:
		 20 21 22 48 49 4d 50 
		We have to be modest, and request only the ports we will be
		interested in using.  */ 

	AR1 = request_region(amosdio_PMIObase+0x48, 2, AMOSDIONAME);  

	// printk("<1>%s(%d): Request region 0x%x length 0x%x = %p\n",
	// 	"amosdio.c", __LINE__, amosdio_PMIObase+0x48, 2, AR1); 

	if(!AR1)
	{
		printk("<1>%s(%d): Failed to Request region 0x%x length 0x%x\n",
			"amosdio.c", __LINE__, amosdio_PMIObase+0x48, 2); 
		return(-1);
	}

	AR2 = request_region(amosdio_PMIObase+0x4c, 2, AMOSDIONAME);  

	// printk("<1>%s(%d): Request region 0x%x length 0x%x = %p\n",
	// 	"amosdio.c", __LINE__, amosdio_PMIObase+0x4c, 2, AR2); 

	if(!AR2)
	{
		printk("<1>%s(%d): Failed to Request region 0x%x length 0x%x\n",
			"amosdio.c", __LINE__, amosdio_PMIObase+0x4c, 2); 
		release_region(amosdio_PMIObase+0x48, 2);
		return(-1);
	}

#if(0)
	/* See if we can read 0x450 anyways */
	emx = inb(0x450);
	printk("<1>%s(%d): Read 0x450 = 0x%x\n", "amosdio.c", __LINE__, emx); 
	outb(emx, 0x450); 
	emx = inb(0x450);
	printk("<1>%s(%d): Read 0x450 = 0x%x should have cleared\n", "amosdio.c", __LINE__, emx); 

	/* We can. We can thus hit on ports 0x420 0x421 0x422 and 
		0x450 without further ado. But with care! Only do the bits 
		that we need to tweak. */
#endif

	amosdio_lo_ucount = 0;
	erx = register_chrdev(AMOSDIOMAJOR, AMOSDIONAME, &amosdio_fops); 
	printk("<1>%s(%d): Register M=%d N=%s = %d\n",
		"amosdio.c", __LINE__, AMOSDIOMAJOR, AMOSDIONAME, erx);


	return(0); 
}

void amosdio_exit(void)
{
	/* Release ports */ 
	release_region(amosdio_PMIObase+0x4c, 2);
	release_region(amosdio_PMIObase+0x48, 2);

	unregister_chrdev(AMOSDIOMAJOR, AMOSDIONAME); 

	printk("<1>%s(%d): Unregister M=%d N=%s\n", 
		"amosdio.c", __LINE__, AMOSDIOMAJOR, AMOSDIONAME); 
}

module_init(amosdio_init);
module_exit(amosdio_exit);


/* Interesting contents of the filp data structure:
	mode_t filp->f_mode; 
	loff_t filp->f_pos;
	unsigned int filp->f_flags;
	struct file_operations *filp->f_op;
	void *filp->private_data; 
	struct dentry *filp->f_dentry;
*/
		
int amosdio_open(struct inode *inode, struct file *filp)
{
	amosdio_lo_ucount ++;
	return(0);
}

int amosdio_release(struct inode *inode, struct file *filp)
{
	amosdio_lo_ucount --;
	return(0);
}


/* Write: just set the output bits, port 0x44d, bits 7 and 6 
	from bits 1 and 0 in the incoming octets */ 

ssize_t amosdio_write(struct file *filp, const char *buff, size_t count, loff_t *offp)
{
	size_t ncco; 
	char wb[3];
	unsigned int itemor, itemand, item;

	ncco = copy_from_user(wb, buff, 1); 
	/* printk("<1>%s(%d): got 0x%x count %d\n", 
		"amosdio.c", __LINE__, wb[0], cco);  */ 

	itemor = wb[0] & 3;
	itemor <<= 6;			/* bits 1 and 0 up to 7 and 6 */

	itemand = itemor;
	itemand |= 0x3F; 		/* ss111111 */
	itemor &= ~0x3F;        /* ss000000 */
	item = inb(0x44d);
	item &= itemand;		/* clears when s is 0 */
	item |= itemor; 		/* sets when s is 1 */ 

	/* printk("<1>%s(%d): got 0x%x OR 0x%x AND 0x%x output 0x%x\n", 
		"amosdio.c", __LINE__, wb[0], itemor, itemand, item); */

	outb(item, 0x44d); 
	// offp ++; 
	return(1); 
} 


/* Read stuff from the AMOSDIO: Read the port and return one byte. 
	This is a byte where the 4 low bits are the 4 
	static input lines, and the high bits are 0x40.
	Thus, this will actually read from @ to O 0x40-0x4F.
*/ 
ssize_t amosdio_read(struct file *filp, char *buff, size_t count, loff_t *offp)
{
	char emits[2];
	unsigned int e1, e2; 
	size_t ncco; 

	/* Read port, copy out, repeat this for as long as we keep asking. */ 
	e1 = inb(0x448);   /* bits 5-4 for GPI 5-4 */ 
	e2 = inb(0x449);   /* bits 7-6 for GPI 3-2 */  
	emits[0] = 0x40 | (e1>>2 & 0x0C) | ( (e2>>6) & 0x03);  
	/* Bits 3-2-1-0 for GPI 5,4,3,2 starts at 0x40: @-O */
	ncco = copy_to_user(buff, emits, 1); 
	// offp ++; 
	return(1);
}


/* ioctl calls.
	ioctl(fd, code, param)

	amosdio_ioctl(inode, file, code, param)

	return 0 for oke, -errornumbers for not ok
	return -EINVAL

 */ 
int amosdio_ioctl(struct inode *inode, struct file *filp, 
	unsigned int code, unsigned long param)
{
	int retval = -ENOIOCTLCMD; 

	switch(code)
	{
	case AMOSDIO_IOCRESET: // __IO(AMOSDIO_MAGIC, 0) 
		retval = 0; 
		break; 

	/* Reset bit 4 in 0x422 to turn the GPI5-SMI action off. */ 
	case AMOSDIO_IOCTGPI5SMIOFF: //  __IO(AMOSDIO_MAGIC, 1) 
	{
		int i22; 
		i22 = inb(0x422);
		printk("<1>%s(%d): found inb(0x%x) = 0x%x;\n", "amosdio.c", __LINE__, 0x422, i22); 
		i22 &= ~0x10; 
		printk("<1>%s(%d): do outb(0%x, 0x%x);\n", "amosdio.c", __LINE__, i22, 0x422); 
		outb(i22, 0x422); 
		retval = 0; 
	}
		break; 

	/* Set bit 4 in 0x422 to turn the GPI5-SMI action on */ 
	case AMOSDIO_IOCTGPI5SMION: // __IO(AMOSDIO_MAGIC, 2) 
	{
		int i22; 
		i22 = inb(0x422);
		printk("<1>%s(%d): found inb(0x%x) = 0x%x;\n", "amosdio.c", __LINE__, 0x422, i22); 
		i22 |= 0x10; 
		printk("<1>%s(%d): do outb(0%x, 0x%x);\n", "amosdio.c", __LINE__, i22, 0x422); 
		outb(i22, 0x422); 
		retval = 0; 
	}
		break; 

	/* Return the change bits values. Copy them out to userspace at *param */ 
	case AMOSDIO_IOCGCHANGES: //  __IOR(AMOSDIO_MAGIC, 3, int) 
	{
		unsigned int i50, i21, i20; 
		unsigned char emits; 

		// Change flag: +0x50 bit 6 set on changes in GPIO2 either direction
		// Change flag: +0x50 bit 7 set on changes in GPIO3 either direction
		// Change flag: +0x21 bit 4 set on change H-L GPI4
		// Change flag: +0x20 bit 4 set on change H-L GPI5

		i20 = inb(0x420);
		printk("<1>%s(%d): found inb(0x%x) = 0x%x;\n", "amosdio.c", __LINE__, 0x420, i20); 
		i21 = inb(0x421);
		printk("<1>%s(%d): found inb(0x%x) = 0x%x;\n", "amosdio.c", __LINE__, 0x421, i21); 
		i50 = inb(0x450);
		printk("<1>%s(%d): found inb(0x%x) = 0x%x;\n", "amosdio.c", __LINE__, 0x450, i50); 

		/* Shifts bits 7-6 down to 1-0, bit 4 down to 2, bit 4 down to 3 */ 
		emits = ((i50>>6) & 0x3) | ((i21 >> 2) & 0x4) | ((i20 >> 1) & 0x8); 
		printk("<1>%s(%d): found emits = 0x%x;\n", "amosdio.c", __LINE__, emits); 

		copy_to_user( (void *) param, &emits, 1); 
		retval = 0; 
	}
		break; 

	case AMOSDIO_IOCSRSTCHANGES: // __IOW(AMOSDIO_MAGIC, 4, int) 
	{
		unsigned int bits;

		copy_from_user(&bits, (void *) param, 1); 

		// Reset Change flag: +0x50 bit 6 set to reset on change flag for GPIO2
		if(bits & AMOSDIO_GPIO2) outb(0x40, 0x450); 

		// Reset Change flag: +0x50 bit 7 set to reset on change flag for GPIO3
		if(bits & AMOSDIO_GPIO3) outb(0x80, 0x450); 

		// Reset Change flag: +0x21 bit 4 set to reset on change flag for GPI4
		if(bits & AMOSDIO_GPI4) outb(0x10, 0x421);

		// Reset Change flag: +0x20 bit 4 set to reset on change flag for GPI5
		if(bits & AMOSDIO_GPI5) outb(0x10, 0x420);
		retval = 0; 
	}
		break; 

	/* No default needed. */ 
	}
	return(retval);
}



